@@ -603,22 +603,30 @@ impl PageData {
     // largest - left on original page.
     // Returns split position
     //
-    fn split(&mut self, new_page: &mut PageData) -> ItemPointer {
+    fn split(&mut self, new_page: &mut PageData, ip: ItemPointer) -> ItemPointer {
         let n_items = self.get_n_items();
         let size = self.get_size();
-        let margin = PAGE_SIZE - size / 2;
-        let mut l: ItemPointer = 0;
         let mut r = n_items;
-        while l < r {
-            let m = (l + r) >> 1;
-            if self.get_offs(m) > margin {
-                // items are allocated from right to left
-                l = m + 1;
-            } else {
-                r = m;
+
+        if ip == r {
+            // Optimization for insert of sequential keys: move all data to new page,
+            // leaving original page empty. It will cause complete filling of B-Tree pages.
+            r -= 1;
+        } else {
+            // Divide page in two approximately equal parts.
+            let margin = PAGE_SIZE - size / 2;
+            let mut l: ItemPointer = 0;
+            while l < r {
+                let m = (l + r) >> 1;
+                if self.get_offs(m) > margin {
+                    // items are allocated from right to left
+                    l = m + 1;
+                } else {
+                    r = m;
+                }
             }
+            debug_assert!(l == r);
         }
-        debug_assert!(l == r);
         // Move first r+1 elements to the new page
         let moved_size = PAGE_SIZE - self.get_offs(r);
 
@@ -734,9 +742,7 @@ impl BufferManager {
                     assert_eq!(self.pages[sync as usize].state, PAGE_DIRTY);
                     if self.pages[sync as usize].access_count == 1 {
                         self.pages[sync as usize].state |= PAGE_SYNCED;
-                        if self.pages[sync as usize].prev != 0 {
-                            self.next_sync = self.pages[sync as usize].prev;
-                        }
+                        self.next_sync = self.pages[sync as usize].prev;
                         let pid = self.pages[sync as usize].pid;
                         next_sync = Some((sync, pid));
                         break;
@@ -771,7 +777,8 @@ impl BufferManager {
         // link to the beginning of dirty list
         if self.dirty_pages != 0 {
             self.pages[self.dirty_pages as usize].prev = id;
-        } else {
+        }
+        if self.next_sync == 0 {
             self.next_sync = id;
         }
         self.pages[id as usize].next = self.dirty_pages;
@@ -1227,7 +1234,7 @@ impl Storage {
                 } else {
                     let mut meta_buf = [0u8; METADATA_SIZE];
                     let len = log.read_at(&mut meta_buf, wal_pos)?;
-                    if len != PAGE_SIZE {
+                    if len != METADATA_SIZE {
                         break;
                     }
                     wal_pos += len as u64;
@@ -1251,6 +1258,7 @@ impl Storage {
                     db.meta_updated = false;
                     db.recovery.recovered_transactions += 1;
                     db.recovery.recovery_end = wal_pos;
+                    crc = 0u32;
                 }
             }
             self.rollback(&mut db)?;
@@ -1342,7 +1350,7 @@ impl Storage {
             // page is full then divide page
             let pin = self.new_page(db)?;
             let mut new_page = self.pool[pin.buf as usize].write().unwrap();
-            let split = page.split(&mut new_page);
+            let split = page.split(&mut new_page, ip);
             let ok = if ip > split {
                 page.insert_item(ip - split - 1, key, value)
             } else {
@@ -1908,6 +1916,16 @@ impl Storage {
         Ok(())
     }
 
+    ///
+    /// Shutdown storage. Unlike close it does't commit delayed transactions, flush data file and truncatate WAL.
+    ///
+    pub fn shutdown(&self) -> Result<()> {
+        let mut db = self.db.write().unwrap();
+        ensure!(db.state == DatabaseState::Opened);
+        db.state = DatabaseState::Closed;
+        Ok(())
+    }
+
     ///
     /// Get recovery status
     ///
